;; Generated package description from buframe.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "buframe" "0.3.0.20260415.112855" "Buffer-local frames" '((emacs "27.1") (timeout "2.1")) :commit "000f49cb5afcdb32ff0695da3b2c8ea207032f4a" :authors '(("Al Haji-Ali" . "abdo.haji.ali@gmail.com")) :maintainer '("Al Haji-Ali" . "abdo.haji.ali@gmail.com") :keywords '("buffer" "frames" "convenience") :url "https://github.com/haji-ali/buframe")
